try:
    import pandas as pd

    _has_pandas = True
    pdSeries = pd.Series
except ImportError:
    _has_pandas = False
    pdSeries = None

try:
    import polars as pl

    _has_polars = True
    plSeries = pl.Series
    plExpr = pl.Expr
except ImportError:
    _has_polars = False
    plSeries = None
    plExpr = None



import typing as t

from aliases.processor import StringProcessor

if t.TYPE_CHECKING:
    from aliases.string import AliasAwareString
    from aliases.dict import AliasAwareDict

_sentinel = object()


class AliasSpace:
    """The alias space stores information on existing aliases

    Parameters
    ----------
    data
        A dictionary where the keys are the preferred forms of strings and the
        values are lists of aliases
    processor
        A string processor that can be used to make certain variants of your
        aliases also count as aliases. When a processed string equals the
        processed form of one of the registered aliases, it is also considered
        to be an alias
    name
        Optional name of this alias space
    """
    def __init__(
        self,
        data: dict[str, list[str]],
        processor: StringProcessor | None = None,
        name: str | None = None,
    ) -> None:
        self._aliases = data
        self._mapper = None
        self.processor = processor if processor is not None else StringProcessor()
        self.name = name

    def add_alias(self, alias: str, representative: str) -> None:
        """Add an alias to the space

        Parameters
        ----------
        alias
            Alias to register
        representative
            Preferred form of the string
        """
        if representative in self._aliases:
            self._aliases[representative].append(alias)
        else:
            self._aliases[representative] = [alias]
            self.mapper[self.processor(representative)] = representative
        self.mapper[self.processor(alias)] = representative

    def create_mapper(self) -> dict[str, str]:
        """Creates a dictionary where the processed form of a registered alias maps to its representative
        """
        result = {
            self.processor(a): rep
            for rep, aliases in self._aliases.items()
            for a in aliases
        }

        # being an alias is reflexive, so we should add a representative as its
        # own alias
        for a in self._aliases:
            result[self.processor(a)] = a

        return result

    @property
    def mapper(self) -> dict[str, str]:
        if self._mapper is None:
            self._mapper = self.create_mapper()
        return self._mapper

    def get_representative(
        self, string: str, missing: t.Any = _sentinel, raise_missing: bool = False
    ) -> t.Any:
        """Get the representative of a given string

        Parameters
        ----------
        string
            The raw input string
        missing
            Value to return when the string is not a registered alias. When this
            argument is not given, the raw input string will be returned back.
        raise_missing
            Whether an error must be raised when the input string is not a registered
            alias
        """
        result = self.mapper.get(self.processor(string), None)
        if result is None:
            if raise_missing:
                raise KeyError(f"{string} not found in {self}")
            if missing is _sentinel:
                return string
            return missing
        return result

    def str(self, data: t.Any) -> "AliasAwareString":
        """Create an AliasAwareString

        Parameters
        ----------
        data
            input string
        """
        from aliases.string import AliasAwareString

        return AliasAwareString(data=data, space=self)

    def dict(self, data: dict[str, t.Any] | None = None, **kwargs) -> "AliasAwareDict":
        """Create an AliasAwareDict

        Parameters
        ----------
        data
            input dictionary
        kwargs
            key, value pairs to add to the dictionary
        """
        from aliases.dict import AliasAwareDict

        return AliasAwareDict(space=self, data=data, **kwargs)

    def map(
        self,
        strings: t.Iterable[str],
        missing: t.Any = _sentinel,
        raise_missing: bool = False,
        return_list: bool = False,
    ) -> map | list:
        """Map an iterable of string to their representatives

        Parameters
        ----------
        strings
            Iterable of strings
        missing
            Value to return when the string is not a registered alias. When this
            argument is not given, the raw input string will be returned back.
        raise_missing
            Whether an error must be raised when the input string is not a registered
            alias
        return_list
            Whether to cast the output map to a list object
        """
        result = map(
            lambda x: self.get_representative(
                x, missing=missing, raise_missing=raise_missing
            ),
            strings,
        )
        if return_list:
            return list(result)
        return result

    def __contains__(self, item: str) -> bool:
        return self.processor(item) in self.mapper

    def __repr__(self) -> str:
        if self.name is None:
            return "AliasSpace"
        return f"AliasSpace {self.name}"



import sys
import os


sys.path.extend([os.path.abspath("../../src")])
import aliases


project = "Aliases"
copyright = "2023, Koen Baak"
author = "Koen Baak"
version = aliases.__version__


extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
]

templates_path = ["_templates"]
exclude_patterns = []
html_theme = "furo"
html_static_path = ["_static"]
html_show_sphinx = False
html_show_copyright = False



import typing as t
from abc import ABC, abstractmethod
from aliases.third_party_classes import (
    _has_polars,
    _has_pandas,
    plSeries,
    plExpr,
    pdSeries,
)

if t.TYPE_CHECKING:
    try:
        import pandas as pd
    except ImportError:
        pass

    try:
        import polars as pl
    except ImportError:
        pass

StringArg: t.TypeAlias = t.Union[str, "pd.Series", "pl.Series", "pl.Expr"]


class StringFunction(ABC):
    """An Abstract Base Class to represent a function that maps strings to strings
    """
    @abstractmethod
    def handle_builtin_str(self, x: str) -> str:
        """Implementation of the function for Python builtin strings"""
        pass

    @abstractmethod
    def handle_pandas_series(self, s: "pd.Series") -> "pd.Series":
        """Implementation of the function for Pandas Series with string values"""
        pass

    @abstractmethod
    def handle_polars_series(self, s: "pl.Series") -> "pl.Series":
        """Implementation of the function for Polars Series with string values"""
        pass

    @abstractmethod
    def handle_polars_expr(self, e: "pl.Expr") -> "pl.Expr":
        """Implementation of the function for Polars Expr"""
        pass

    def __call__(self, x: StringArg) -> StringArg:
        """Call this StringFunction

        This function will dispatch to one of the implementations depending
        on the type of the input argument

        Parameters
        ----------
        x
            Input String Argument
        """
        if isinstance(x, str):
            return self.handle_builtin_str(x)
        elif _has_pandas and isinstance(x, pdSeries):
            return self.handle_pandas_series(x)
        elif _has_polars and isinstance(x, plSeries):
            return self.handle_polars_series(x)
        elif _has_polars and isinstance(x, plExpr):
            return self.handle_polars_expr(x)
        raise TypeError


class StringIdentity(StringFunction):
    def handle_builtin_str(self, x: str) -> str:
        return x

    def handle_pandas_series(self, s: "pd.Series") -> "pd.Series":
        return s

    def handle_polars_expr(self, e: "pl.Expr") -> "pl.Expr":
        return e

    def handle_polars_series(self, s: "pl.Series") -> "pl.Series":
        return s


class ToLowerCase(StringFunction):
    def handle_builtin_str(self, x: str) -> str:
        return x.lower()

    def handle_pandas_series(self, s: "pd.Series") -> "pd.Series":
        return s.str.lower()

    def handle_polars_expr(self, e: "pl.Expr") -> "pl.Expr":
        return e.str.to_lowercase()

    def handle_polars_series(self, s: "pl.Series") -> "pl.Series":
        return s.str.to_lowercase()


class RightStrip(StringFunction):
    def handle_builtin_str(self, x: str) -> str:
        return x.rstrip()

    def handle_pandas_series(self, s: "pd.Series") -> "pd.Series":
        return s.str.rstrip()

    def handle_polars_expr(self, e: "pl.Expr") -> "pl.Expr":
        return e.str.rstrip()

    def handle_polars_series(self, s: "pl.Series") -> "pl.Series":
        return s.str.rstrip()



import typing as t
from aliases.string_function import (
    StringFunction,
    StringArg,
    StringIdentity,
    ToLowerCase,
    RightStrip,
)


class StringProcessor:
    def __init__(self, f: StringFunction | None = None) -> None:
        self.f = f if f is not None else StringIdentity()

    def __call__(self, x: StringArg) -> StringArg:
        return self.f(x)

    def compose(
        self, other: t.Union[StringFunction, "StringProcessor"]
    ) -> "StringProcessor":
        current = self.f
        f = other if not isinstance(other, StringProcessor) else other.f
        processor = StringProcessor()
        processor.f = lambda x: f(current(x))
        return processor

    def lower(self) -> "StringProcessor":
        return self.compose(ToLowerCase())

    def rstrip(self) -> "StringProcessor":
        return self.compose(RightStrip())



import typing as t

from aliases.space import AliasSpace, _sentinel
import polars as pl


@pl.api.register_series_namespace("aliases")
class PolarsSeriesAliasAccessor:
    def __init__(self, s: pl.Series) -> None:
        self._s = s

    def representative(
        self,
        space: AliasSpace,
        missing: t.Any = _sentinel,
        raise_missing: bool = False,
    ) -> pl.Series:
        result: pl.Series = space.processor(self._s)
        if raise_missing:
            default = None
        elif missing is _sentinel:
            default = pl.first()
        else:
            default = missing
        result = result.map_dict(remapping=space.mapper, default=default)
        if raise_missing:
            is_missing = result.is_null()
            if is_missing.any():
                raise KeyError(
                    f"{self._s.filter(is_missing).to_list()} not found in {space}"
                )
        return result


@pl.api.register_expr_namespace("aliases")
class PolarsExprAliasAccessor:
    def __init__(self, expr: pl.Expr) -> None:
        self._expr = expr

    def representative(self, space: AliasSpace, missing: t.Any = _sentinel) -> pl.Expr:
        result: pl.Expr = space.processor(self._expr)

        if missing is _sentinel:
            default = pl.first()
        else:
            default = missing
        return result.map_dict(remapping=space.mapper, default=default)


@pl.api.register_dataframe_namespace("aliases")
class PolarsDataFrameAliasAccessor:
    def __init__(self, df: pl.DataFrame) -> None:
        self._df = df

    def representative(
        self,
        space: AliasSpace,
        columns: list[str] | None = None,
        missing: t.Any = _sentinel,
        raise_missing: bool = False,
    ) -> pl.DataFrame:
        cols = columns if columns is not None else pl.Utf8
        missing = missing if not raise_missing else None
        result = self._df.with_columns(
            pl.col(cols).aliases.representative(space=space, missing=missing)
        )

        if raise_missing:
            if result.select(pl.col(cols)).null_count().sum(axis=1) > 0:
                raise KeyError("Some element in the DataFrame could not be found")
        return result


@pl.api.register_lazyframe_namespace("aliases")
class PolarsLazyFrameAliasAccessor:
    def __init__(self, ldf: pl.LazyFrame) -> None:
        self._ldf = ldf

    def representative(
        self,
        space: AliasSpace,
        columns: list[str] | None = None,
        missing: t.Any = _sentinel,
    ) -> pl.LazyFrame:
        cols = columns if columns is not None else pl.Utf8
        return self._ldf.with_columns(
            pl.col(cols).aliases.representative(space=space, missing=missing)
        )



__version__ = "0.6.2"

from aliases.space import AliasSpace
from aliases.string import AliasAwareString
from aliases.dict import AliasAwareDict
from aliases.processor import StringProcessor

try:
    from aliases.pandas_accessor import (
        PandasSeriesAliasAccessor,
        PandasDataFrameAliasAccessor,
    )
except ImportError:
    pass

try:
    from aliases.polars_accessor import (
        PolarsSeriesAliasAccessor,
        PolarsExprAliasAccessor,
        PolarsDataFrameAliasAccessor,
        PolarsLazyFrameAliasAccessor,
    )
except ImportError:
    pass

__all__ = [AliasSpace, AliasAwareDict, AliasAwareString, StringProcessor]
